local noop = function() end
local animation_path = "canodumb.animation"

-- cursor.lua
local Cursor = {}

local function spawn_barrel_smoke(canodumb)
  local field = canodumb:field()
  local tile = canodumb:current_tile()

  local artifact = Artifact.new()
  artifact:set_facing(canodumb:facing())
  artifact:sprite():set_layer(-1)
  artifact:set_texture(canodumb:texture(), true)

  if canodumb:facing() == Direction.Left then
    artifact:set_offset(-40 * 0.5, -65 * 0.5)
  else
    artifact:set_offset(40 * 0.5, -65 * 0.5)
  end

  local anim = artifact:animation()
  anim:load(animation_path)
  anim:set_state("SMOKE")
  anim:on_complete(function()
    artifact:erase()
  end)

  field:spawn(artifact, tile:x(), tile:y())
end

local function attack_tile(canodumb, attack_spell, tile)
  if not tile then
    attack_spell:erase()
    return
  end

  -- set the cursor up for damage
  attack_spell:set_hit_props(HitProps.new(
    canodumb._attack,
    Hit.Flinch | Hit.Impact | Hit.Flash,
    Element.Cursor,
    canodumb:context(),
    Drag.new()
  ))
  tile:attack_entities(attack_spell)
end

local function attack(canodumb, cursor, tile)
  Resources.play_audio(canodumb._cannon_sfx)

  spawn_barrel_smoke(canodumb)

  local cursor_anim = cursor.spell:animation()
  cursor_anim:set_state("CURSOR_SMOKE")
  cursor_anim:set_playback(Playback.Once)
  cursor_anim:on_complete(function()
    cursor:erase()
  end)

  local canodumb_anim = canodumb:animation()
  canodumb_anim:set_state(canodumb._shoot_state)
  canodumb_anim:on_complete(function()
    canodumb_anim:set_state(canodumb._idle_state)
  end)

  -- create a new attack spell to deal damage
  -- we can't deal damage to a target if we've hit them and they haven't moved
  local attack_spell = Spell.new(canodumb:team())

  local tile = canodumb._cursor.spell.tile
  local direction = canodumb:facing()
  local delete_cooldown = 1
  attack_spell.on_update_func = function(self)
    attack_tile(canodumb, attack_spell, tile)
    if delete_cooldown <= 0 then
      self:erase()
    else
      delete_cooldown = delete_cooldown - 1
    end
  end

  attack_spell.on_attack_func = function()
    -- hit something, dont hit anything more
    attack_spell:erase()
  end

  -- attach the spell so attacks can register
  canodumb:field():spawn(attack_spell, tile)
end

local function begin_attack(canodumb, cursor)
  -- stop the cursor from scanning for players
  cursor.spell.on_update_func = noop
  cursor.spell.on_attack_func = noop

  local cursor_anim = cursor.spell:animation()
  cursor_anim:set_state(canodumb._cursor_shoot_state)
  cursor_anim:set_playback(Playback.Once)
  cursor_anim:on_complete(function()
    canodumb._should_attack = true
  end)
end

local function spawn_cursor(cursor, canodumb, tile, target)
  local spell = Spell.new(canodumb:team())
  spell:set_texture(canodumb:texture(), true)
  spell:sprite():set_layer(-1)
  spell:set_hit_props(
    HitProps.new(
      0,
      Hit.None,
      Element.None,
      canodumb:context(),
      Drag.new()
    )
  )

  spell.did_collide = false;

  local anim = spell:animation()
  anim:load(animation_path)
  anim:set_state(canodumb._cursor_state)
  anim:set_playback(Playback.Once)

  spell.slide_started = false

  local field = canodumb:field()
  field:spawn(spell, tile)
  local cursor_delete_cooldown = 1800
  spell.on_update_func = function(self)
    if cursor_delete_cooldown <= 0 then
      cursor:erase()
    else
      cursor_delete_cooldown = cursor_delete_cooldown - 1
    end
    if target ~= nil and not target:deleted() then
      spell.tile = target:current_tile()
    else
      spell.tile = spell:get_tile(spell:facing(), 1)
    end
    if tile then
      if spell:current_tile():is_edge() and spell.slide_started then
        spell:delete()
      end
      if not spell:is_sliding() then
        local ref = spell
        spell:slide(spell.tile, canodumb._frames_per_cursor_movement, function()
          ref.slide_started = true
        end)
      end
    else
      cursor:erase()
    end
  end

  spell.on_collision_func = function()
    if spell.did_collide == false then
      spell.did_collide = true;
      begin_attack(canodumb, cursor)
    end
  end

  spell.can_move_to_func = function(tile)
    return true
  end

  return spell
end

function Cursor:new(canodumb, tile, target)
  local cursor = {
    canodumb = canodumb,
    spell = nil,
    remaining_frames = canodumb._frames_per_cursor_movement
  }

  setmetatable(cursor, self)
  self.__index = self

  cursor.spell = spawn_cursor(cursor, canodumb, tile, target)

  return cursor
end

function Cursor:erase()
  self.spell:erase()
  self.canodumb._cursor = nil
end

-- entry.lua

local target_update, idle_update

target_update = function(canodumb)
  if canodumb._cursor then
    local spell = canodumb._cursor.spell

    spell:current_tile():attack_entities(spell)

    if canodumb._should_attack then
      canodumb._should_attack = false
      attack(canodumb, canodumb._cursor)
    end
  else
    canodumb.on_update_func = idle_update
  end
end

idle_update = function(canodumb)
  local y = canodumb:current_tile():y()
  local team = canodumb:team()

  local field = canodumb:field()
  local targets = field:find_characters(function(c)
    -- same row, different team
    return c:team() ~= team and c:current_tile():y() == y
  end)

  if #targets == 0 then return end -- no target

  -- found a target, spawn a cursor and change state
  local cursor_tile = canodumb:get_tile(canodumb:facing(), 1)
  canodumb._cursor = Cursor:new(canodumb, cursor_tile, targets[1])
  canodumb.on_update_func = target_update
end

local on_delete_func = function(canodumb)
  if canodumb._cursor then
    canodumb._cursor:erase()
  end
  canodumb:default_character_delete()
end

function character_init(canodumb)
  canodumb._cannon_sfx = Resources.load_audio("cannon.ogg")

  -- private variables
  canodumb._frames_per_cursor_movement = 60
  canodumb._cursor = nil
  canodumb._idle_state = "IDLE_1"
  canodumb._shoot_state = "SHOOT_1"
  canodumb._cursor_state = "CURSOR"
  canodumb._cursor_shoot_state = "CURSOR_SHOOT"
  canodumb._attack = 100
  canodumb._should_attack = false

  -- meta
  canodumb:set_name("Canosmart")
  canodumb:set_height(55)
  canodumb:set_health(500)

  canodumb:set_texture(Resources.load_texture("canodumb.png"))

  local anim = canodumb:animation()
  anim:load(animation_path)
  anim:set_state(canodumb._idle_state)
  anim:set_playback(Playback.Once)

  -- setup defense rules
  canodumb.defense = DefenseVirusBody.new() -- lua owns this need to keep it alive
  canodumb:add_defense_rule(canodumb.defense)

  -- setup event hanlders
  canodumb.on_update_func = idle_update
  canodumb.on_battle_start_func = noop
  canodumb.on_battle_end_func = noop
  canodumb.on_spawn_func = noop
  canodumb.on_delete_func = on_delete_func
end
